/*
 *  $ReleaseID$
 *
 *  $Id$
 */

/*
 *      This file contains the code that interfaces the kernel
 *      for the ARVID VCR-tape driver for Linux.
 */

#include <linux/module.h>
#include <linux/version.h>
#include <linux/errno.h>
#include <linux/fs.h>

#include <asm/segment.h>
#include <linux/kernel.h>
#include <linux/signal.h>
#include <linux/major.h>
#include <linux/malloc.h>
#include <asm/dma.h>

#include "arvid.h"
#include "arv-hlp.h"
#include "arv-data.h"
#include "arv-io.h"

#include "tracing.h"
#include "arvid-ki.h"

static int arvid_unit,arvid_failure;
static int old_sigmask;

static int arvid_open(struct inode *ino, struct file *filep);
static void arvid_close(struct inode *ino, struct file *filep);
static int arvid_ioctl(struct inode *ino, struct file *filep,
		       unsigned int command, unsigned long arg);
static int arvid_read(struct inode *ino, struct file *fp, char *buff,
		      int req_len);
static int arvid_write(struct inode *ino, struct file *fp, const char *buff,
		       int req_len);

static struct file_operations arvid_cdev =
{
	NULL,			/* lseek */
	arvid_read,		/* read */
	arvid_write,		/* write */
	NULL,			/* readdir */
	NULL,			/* select */
	arvid_ioctl,		/* ioctl */
	NULL,			/* mmap */
	arvid_open,		/* open */
	arvid_close,		/* release */
	NULL,			/* fsync */
};

#ifdef MODULE
#define arvid_init init_module
#endif

static void __unregister(void) {
  TRACE_FUN(5, "__unregister");
  if (unregister_chrdev(ARVID_DEVMAJOR, "arvid") != 0) {
    TRACE(3, "unregister_chrdev failed");
  } else {
    TRACE(3, "unregister_chrdev successful");
  }
  TRACE_EXIT;
}

int arvid_init(void)
{
	int err;
	TRACE_FUN(5, "arvid_init");
#ifdef MODULE
	printk(KERN_INFO "arvid-0.0 ????97\n"
	       KERN_INFO " (c) 1996-1997 Dmitry Prokhorov (dip@sunnit.ifmo.ru)\n"
	       KERN_INFO " ARVID driver for ARVID 1031..1052 controllers\n"
	       KERN_INFO " Compiled for kernel version %s"
#ifdef MODVERSIONS
	       " with versioned symbols"
#endif
	       "\n", kernel_version);
	printk(KERN_INFO "iobase=0x%03X, irq=%d\n", arvparam[0], arvparam[1]);
#else /* !MODULE */
	/* print a short no-nonsense boot message */
	printk("arvid-0.0 ????97 for Linux 2.0.27\n");
	printk(KERN_INFO "iobase=0x%03X, irq=%d\n", ARVID_IOBASE, ARVID_IRQ);
#endif				/* MODULE */
	TRACE(3, "installing arvid driver...");
	if (register_chrdev(ARVID_DEVMAJOR, "arvid", &arvid_cdev)) {
		TRACE(1, "register_chrdev failed");
		TRACE_EXIT;
		return -EIO;
	}

	TRACEx1(3, "arvid_init @ 0x%p", arvid_init);

	arvid.busy=0;
	arvid.busy_stat=0;
	arvid.stat_read=0;

	err = _arvid_init();
	if(arvid.model == UNDETECTED) {
	  printk(KERN_INFO "ARVID model undetected\n");
	  TRACE_EXIT;
	  __unregister();
	  return -EIO;
	} else 
	  printk(KERN_INFO "ARVID detected, model: Arvid-%i\n", arvid.modidx);

	if(err != ARVERR_OK) {
	  TRACE(1, "arvid initialisation failed");
	  TRACE_EXIT;
	  __unregister();
	  return -EIO;
	}

	err = _arvid_get();
	if(err != 0) {
		TRACE(1, "interrupt request failed");
		TRACE_EXIT;
		__unregister();
		return -EIO;
	}

	_arvid_reset();

	arvid.busy = 0;
	arvid_failure = 1;	/* inhibit any operation but open */

	TRACE_EXIT;
#ifdef MODULE
	register_symtab(0);	/* remove global arvid symbols */
#endif
	return 0;
}


#ifdef MODULE
/*      Called by modules package when removing the driver
 */
void cleanup_module(void)
{
/*	int n; */
	TRACE_FUN(5, "cleanup_module");
	__unregister();
	_arvid_reset();
	_arvid_free();
	TRACE_EXIT;
}
#endif				/* MODULE */


/*      Open arvid device
 */
static int arvid_open(struct inode *ino, struct file *filep)
{
	TRACE_FUN(4, "arvid_open");
	int result=0;
/*  	MOD_INC_USE_COUNT; */	/* lock module in memory */

	TRACEi(5, "called for minor", MINOR(ino->i_rdev));

        if (MINOR(ino->i_rdev) < ARVID_DEVSTAT && MINOR(ino->i_rdev) > ARVID_MAXUNIT) {
		TRACE(1, "failed: illegal unit nr");
		/*MOD_DEC_USE_COUNT;*/
		TRACE_EXIT;
		return -ENXIO;
	}

	if(MINOR(ino->i_rdev) >= ARVID_DEVSTAT && arvid.busy_stat) {
		TRACE(1, "failed: status already busy");
		/*MOD_DEC_USE_COUNT;*/	/* unlock module in memory */
		TRACE_EXIT;
		return -EBUSY;
	}

	if (MINOR(ino->i_rdev) < ARVID_DEVSTAT && arvid.busy) {
		TRACE(1, "failed: already busy");
		/*MOD_DEC_USE_COUNT;*/	/* unlock module in memory */
		TRACE_EXIT;
		return -EBUSY;
	}
/*
        if (arvid_unit == -1 || ARVID_UNIT != (MINOR(ino->i_rdev) & 3)) {
		arvid_init_driver();
	}
*/
	if(MINOR(ino->i_rdev) >= ARVID_DEVSTAT) {
	  arvid.busy_stat = 1;
	  arvid.stat_read=0;
	  TRACE_EXIT;
	  return 0;
	}

	arvid_unit = MINOR(ino->i_rdev);

	arvid_failure = 0;	/* allow tape operations */
	old_sigmask = current->blocked;
	current->blocked = _BLOCK_ALL;

	_arvid_reset();

	result = _arvid_open();

	if (result < 0) {
		TRACE(1, "_arvid_open failed");
		current->blocked = old_sigmask;		/* restore mask */
		/*MOD_DEC_USE_COUNT;*/	/* unlock module in memory */
		TRACE_EXIT;
		return result;
	} else {
		arvid.busy = 1;
		/*  Mask signals that will disturb proper operation of the
		 *  program that is calling.
		 */
		current->blocked = old_sigmask | _DO_BLOCK;
		TRACE_EXIT;
		return 0;
	}
}

/*      Close arvid device
 */
static void arvid_close(struct inode *ino, struct file *filep)
{
	TRACE_FUN(4, "arvid_close");
	int result=0;

	if ((!arvid.busy && !arvid.busy_stat) ||
	    (!arvid.busy && MINOR(ino->i_rdev) < ARVID_DEVSTAT) ||
	    (!arvid.busy_stat && MINOR(ino->i_rdev) >= ARVID_DEVSTAT)) {
	       TRACE(1, "failed: not busy or wrong unit");
	       TRACE_EXIT;
	       return;		/* keep arvid.busy !(?) */
	}

	if(MINOR(ino->i_rdev) >= ARVID_DEVSTAT) {
	  arvid.busy_stat = 0;
	  TRACE_EXIT;
	  /*  MOD_DEC_USE_COUNT; */
	  return;
	}
	current->blocked = _BLOCK_ALL;

	result = _arvid_close();

	if (result < 0) {
		TRACE(1, "_arvid_close failed");
	}

	_arvid_reset();

	arvid_failure = 1;	/* inhibit any operation but open */
	arvid.busy = 0;
	current->blocked = old_sigmask;		/* restore before open state */
	TRACE_EXIT;
/*	MOD_DEC_USE_COUNT; */	/* unlock module in memory */
}

/*      Ioctl for arvid device
 */
static int arvid_ioctl(struct inode *ino, struct file *filep,
		       unsigned int command, unsigned long arg)
{
	TRACE_FUN(4, "arvid_ioctl");
	int result = -EIO;
	int old_sigmask;

	if (arvid.busy_stat && MINOR(ino->i_rdev) == ARVID_DEVSTAT) {
	  TRACE_EXIT;
	  return 0;
	}
	if (!arvid.busy || MINOR(ino->i_rdev) != arvid_unit || arvid_failure) {
		TRACE(1, "failed: not busy, failure or wrong unit");
		TRACE_EXIT;
		return -EIO;
	}
	old_sigmask = current->blocked;		/* save mask */
	current->blocked = _BLOCK_ALL;
	/* This will work as long as sizeof( void*) == sizeof( long)
	 */
/* <-- here
	result = _arvid_ioctl(command, (void *) arg);
*/
	current->blocked = old_sigmask;		/* restore mask */
	TRACE_EXIT;
	return result;
}

/*      Read from tape device
 */
static int arvid_read(struct inode *ino, struct file *fp, char *buff, int req_len)
{
	TRACE_FUN(5, "arvid_read");
	int result = -EIO;
	int old_sigmask;

	TRACEi(5, "called with count:", req_len);
	if (arvid.busy_stat && MINOR(ino->i_rdev) == ARVID_DEVSTAT) {
	  if(arvid.stat_read) return 0;
	  else {
	    arvid.stat_read=1;
	    return _arvid_status(buff, req_len);
	    TRACE_EXIT;
	  }
	}
	if (!arvid.busy || MINOR(ino->i_rdev) != arvid_unit || arvid_failure) {
		TRACE(1, "failed: not busy, failure or wrong unit");
		TRACE_EXIT;
		return -EIO;
	}
	old_sigmask = current->blocked;		/* save mask */
	current->blocked = _BLOCK_ALL;

	result = _arvid_read(buff, req_len);

	TRACEi(7, "return with count:", result);
	current->blocked = old_sigmask;		/* restore mask */
	TRACE_EXIT;
	return result;
}

/*      Write to tape device
 */
static int arvid_write(struct inode *ino, struct file *fp, const char *buff, int req_len)
{
	TRACE_FUN(8, "arvid_write");
	int result = -EIO;
	int old_sigmask;

	TRACEi(5, "called with count:", req_len);
	if (arvid.busy_stat && MINOR(ino->i_rdev) == ARVID_DEVSTAT) {
	  TRACE_EXIT;
	  return 0;
	}
	if (!arvid.busy || MINOR(ino->i_rdev) != arvid_unit || arvid_failure) {
		TRACE(1, "failed: not busy, failure or wrong unit");
		TRACE_EXIT;
		return -EIO;
	}
	old_sigmask = current->blocked;		/* save mask */
	current->blocked = _BLOCK_ALL;

	result = _arvid_write(buff, req_len);

	TRACEi(7, "return with count:", result);
	current->blocked = old_sigmask;		/* restore mask */
	TRACE_EXIT;
	return result;
}
